from can_id_sheet import CANFrame, VehicleSegment, Frequency
from periodic_send_thread import PeriodicSendingTask
from multiprocessing import Process
from scapy.sendrecv import *
from scapy.layers.l2 import Ether
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.exceptions import InvalidSignature


# Load public key from file
with open("public_key.pem", "rb") as key_file:
    public_key = serialization.load_pem_public_key(
        key_file.read(),
    )


# Function to decrypt and verify data using AES-GCM
def decrypt_and_verify(key, encrypted_data, associated_data=None):
    aesgcm = AESGCM(key)
    nonce = encrypted_data[:12]
    ciphertext = encrypted_data[12:]
    decrypted_data = aesgcm.decrypt(nonce, ciphertext, associated_data)
    return decrypted_data


def verify_signature(pub_key, packet):
    payload = bytes(packet["Raw"].load[:-256])
    signature = bytes(packet["Raw"].load[-256:])
    try:
        pub_key.verify(
            signature,
            payload,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        return True
    except InvalidSignature:
        return False


class GenericECU(Process):
    def __init__(self,
                 protocol: str,
                 interface: str,
                 tx_socket,  # type: SuperSocket
                 rx_socket,  # type: SuperSocket
                 vehicle_segment: VehicleSegment
                 ):
        self.msg_count = 0

        def sniff_function(pkt):
            self.msg_count += 1
            # print(pkt.show())

        def eth_decryption_sniff_function(pkt):
            # print("Received encrypted Ether packet:")
            # print(pkt.show())
            self.msg_count += 1
            received_pkt_payload = pkt.load
            # Access src and dst MAC addresses
            src_mac = pkt.src
            dst_mac = pkt.dst
            src_mac_bytes = bytes.fromhex(src_mac.split(':')[0])
            dst_mac_bytes = bytes.fromhex(dst_mac.split(':')[0])
            assoc_data = src_mac_bytes + dst_mac_bytes
            # Decrypt the received packet
            decrypted_packet = decrypt_and_verify(key, received_pkt_payload, assoc_data)
            received_ether_packet = Ether(src=pkt.src, dst=pkt.dst)/decrypted_packet
            # print("Received decrypted Ether packet:")
            # print(received_ether_packet.show())

        def eth_verify_sign_sniff_function(pkt):
            # print("Received Ether signed packet:")
            # print(pkt.show())
            self.msg_count += 1
            # Verify the signature using the public key
            # if verify_signature(public_key, pkt):
            #     print("Signature verification successful.")
            # else:
            #     print("Signature verification failed.")

        Process.__init__(self)
        self.protocol = protocol
        self.interface = interface
        self.tx_socket = tx_socket
        self.rx_socket = rx_socket
        self.frames_to_send = [frame for frame in CANFrame if frame.source == vehicle_segment]
        self.name = vehicle_segment.name

        ecu_work_description = f'Work description for {self.name} ECU:\n'
        if protocol == "can":  # can
            tasks = [AsyncSniffer(opened_socket=self.rx_socket, prn=sniff_function)]
        else:  # ethernet
            # Load the decryption key
            password = b"supersecretpassword"
            salt = bytes(10)
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                iterations=100000,
                salt=salt,
                length=32
            )
            key = kdf.derive(password)
            tasks = [AsyncSniffer(iface=self.rx_socket.getsockname()[0],
                                  # Use normal sniff function
                                  # prn=sniff_function,
                                  # Use decryption
                                  # prn=eth_decryption_sniff_function,
                                  # Use verify signature
                                  prn=eth_verify_sign_sniff_function,
                                  filter=f"ether dst host 0{vehicle_segment.value}:00:00:00:00:00")]

        # split frames of this ECU by sending frequency and schedule them on separate cyclic threads
        for freq in Frequency:
            frames_for_freq = list(filter(lambda x: x.interval == freq, self.frames_to_send))
            if not frames_for_freq:
                continue
            tasks.append(PeriodicSendingTask(self.protocol, interface=interface, socket=self.tx_socket, frames=frames_for_freq, interval_s=freq.value))
            ecu_work_description += f' - Created cyclic thread which will send {frames_for_freq} every {freq.value}s.\n'

        self.tasks = tasks

        print(ecu_work_description)

    def run(self):
        print(f'Hello from ECU with segment: {self.name} and socket: {self.tx_socket}.')

        for t in self.tasks:
            t.start()

        time.sleep(60)

        for t in self.tasks:
            t.join(timeout=5)

        print(f'Message count {self.msg_count} from ECU with segment: {self.name}')
