from can_id_sheet import CANFrame
from scapy.layers.can import *
from scapy.layers.l2 import Ether
from scapy.all import *
from scapy.sendrecv import send
import time
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.asymmetric import padding

NANOSECONDS_IN_SECOND: int = 1_000_000_000


# Load private key from file
with open("private_key.pem", "rb") as key_file:
    private_key = serialization.load_pem_private_key(
        key_file.read(),
        password=None,  # No password protection
    )


# Function to sign the data
def sign_packet(pkey, packet_data):
    signature = pkey.sign(
        packet_data,
        padding.PSS(
            mgf=padding.MGF1(hashes.SHA256()),
            salt_length=padding.PSS.MAX_LENGTH
        ),
        hashes.SHA256()
    )
    return signature


# Function to encrypt and authenticate data using AES-GCM
def encrypt_and_authenticate(key, data, associated_data=None):
    aesgcm = AESGCM(key)
    nonce = os.urandom(12)  # Generate a random nonce
    ciphertext = aesgcm.encrypt(nonce, data, associated_data)  # Encrypt the data
    return nonce + ciphertext


class PeriodicSendingTask(Thread):
    def __init__(self,
                 protocol: str,
                 interface: str,
                 socket,  # type: SuperSocket,
                 frames: list[CANFrame],
                 interval_s: float):
        super().__init__()
        self.daemon = True
        self.stopped = False    # maybe add a way to gracefully stop
        self.protocol = protocol
        self.interface = interface  # ethernet network interface
        self.socket = socket
        self.frames = frames
        self.interval_s = interval_s
        self.interval_ns = int(round(interval_s * NANOSECONDS_IN_SECOND))
        if protocol == "can":
            self.packets = [CAN(identifier=f.value, data=b'01020304') for f in frames]
        elif protocol == "eth":
            payload = b"x" * 64
            self.packets = [Ether(src=f.source.value, dst=f.execution.value)/payload for f in self.frames]

    def run(self) -> None:
        msg_due_time_ns = time.perf_counter_ns()
        # Load the encryption/decryption key (must be kept secure!)
        password = b"supersecretpassword"
        salt = bytes(10)
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            iterations=100000,
            salt=salt,
            length=32
        )
        key = kdf.derive(password)

        while not self.stopped:
            if self.protocol == "can":
                # CAN packet sending
                send(self.packets, verbose=0, socket=self.socket)
                # print(f'Sent {len(self.packets)} CAN packets at {time.time()}, frequency = {self.interval_s}s')
            elif self.protocol == "eth":
                # encrypt each packet payload and create a list of encrypted packets
                encrypted_packets = []
                for p in self.packets:
                #     # Encryption
                #     src_mac_bytes = p.src.to_bytes(1, 'big')
                #     dst_mac_bytes = p.dst.to_bytes(1, 'big')
                #     assoc_data = src_mac_bytes + dst_mac_bytes
                #     encrypted_payload = encrypt_and_authenticate(key, p.load, assoc_data)
                #     encrypted_packet = Ether(src=p.src, dst=p.dst)/encrypted_payload
                #     encrypted_packets.append(encrypted_packet)

                    # Sign data
                    signature = sign_packet(private_key, p.load)
                    signed_packet = p / signature
                    encrypted_packets.append(signed_packet)
                # ethernet packet sending
                sendp(encrypted_packets, verbose=0, iface=self.socket.getsockname()[0])
                # For debug
                # for p in encrypted_packets:
                #     p.show()
                # print(f'Sent {len(self.packets)} Ethernet packets at {time.time()}, frequency = {self.interval_s}s')

            msg_due_time_ns += self.interval_ns
            # Compensate for the time it takes to send the message
            delay_ns = msg_due_time_ns - time.perf_counter_ns()
            if delay_ns > 0:
                time.sleep(delay_ns / NANOSECONDS_IN_SECOND)
